@@ -13,11 +13,14 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from http import client as http_client
 import logging
+import re
 import time
 from urllib import parse as urlparse
 
 import requests
+from urllib3.exceptions import InsecureRequestWarning
 
 from sushy import exceptions
 from sushy.taskmonitor import TaskMonitor
@@ -26,21 +29,19 @@ from sushy import utils
 LOG = logging.getLogger(__name__)
 
 
-_SERVER_SIDE_RETRIES = 5
-_SERVER_SIDE_RETRY_DELAY = 3
-
-
 class Connector(object):
 
     def __init__(
             self, url, username=None, password=None, verify=True,
-            response_callback=None):
+            response_callback=None, server_side_retries=0,
+            server_side_retries_delay=0):
         self._url = url
         self._verify = verify
         self._session = requests.Session()
-        self._session.verify = self._verify
         self._response_callback = response_callback
         self._auth = None
+        self._server_side_retries = server_side_retries
+        self._server_side_retries_delay = server_side_retries_delay
 
         # NOTE(TheJulia): In order to help prevent recursive post operations
         # by allowing us to understand that we should stop authentication.
@@ -67,6 +68,13 @@ class Connector(object):
                         'be removed in the future')
             self.set_http_basic_auth(username, password)
 
+        if not self._verify:
+            # As the user specifically needs to opt out of certificate
+            # validation the user is aware of the security implications
+            # and does not need to be overwhelmed by the urllib3 warnings
+            requests.packages.urllib3.disable_warnings(
+                category=InsecureRequestWarning)
+
     def set_auth(self, auth):
         """Sets the authentication mechanism for our connector."""
         self._auth = auth
@@ -84,8 +92,17 @@ class Connector(object):
         """Close this connector and the associated HTTP session."""
         self._session.close()
 
+    def check_retry_on_exception(self, exception_msg):
+        """Checks whether retry on exception is required."""
+        if ('SYS518' in str(exception_msg)):
+            LOG.debug('iDRAC is not yet ready after previous operation. '
+                      'Error: %(err)s', {'err': str(exception_msg)})
+            return True
+        else:
+            return False
+
     def _op(self, method, path='', data=None, headers=None, blocking=False,
-            timeout=60, server_side_retries=_SERVER_SIDE_RETRIES,
+            timeout=60, server_side_retries_left=None,
             **extra_session_req_kwargs):
         """Generic RESTful request handler.
 
@@ -96,6 +113,8 @@ class Connector(object):
         :param headers: Optional dictionary of headers.
         :param blocking: Whether to block for asynchronous operations.
         :param timeout: Max time in seconds to wait for blocking async call.
+        :param server_side_retries_left: Remaining retries. If not provided
+            will use limit provided by instance's server_side_retries
         :param extra_session_req_kwargs: Optional keyword argument to pass
          requests library arguments which would pass on to requests session
          object.
@@ -103,6 +122,10 @@ class Connector(object):
         :raises: ConnectionError
         :raises: HTTPError
         """
+
+        if server_side_retries_left is None:
+            server_side_retries_left = self._server_side_retries
+
         url = path if urlparse.urlparse(path).netloc else urlparse.urljoin(
             self._url, path)
         headers = headers or {}
@@ -124,6 +147,7 @@ class Connector(object):
         try:
             response = self._session.request(method, url, json=data,
                                              headers=headers,
+                                             verify=self._verify,
                                              **extra_session_req_kwargs)
         except requests.exceptions.RequestException as e:
             # Capture any general exception by looking for the parent
@@ -181,6 +205,7 @@ class Connector(object):
                     response = self._session.request(
                         method, url, json=data,
                         headers=headers,
+                        verify=self._verify,
                         **extra_session_req_kwargs)
                 except exceptions.HTTPError as retry_exc:
                     LOG.error("Failure occured while attempting to retry "
@@ -197,17 +222,23 @@ class Connector(object):
                               "%s", e.message)
                 raise
         except exceptions.ServerSideError as e:
-            if method.lower() != 'get' or server_side_retries <= 0:
-                raise
-            else:
+            if ((method.lower() == 'get'
+                or self.check_retry_on_exception(e.message))
+                    and server_side_retries_left > 0):
                 LOG.warning('Got server side error %s in response to a '
-                            'GET request, retrying after %d seconds',
-                            e, _SERVER_SIDE_RETRY_DELAY)
-                time.sleep(_SERVER_SIDE_RETRY_DELAY)
-                return self._op(method, path, data=data, headers=headers,
-                                blocking=blocking, timeout=timeout,
-                                server_side_retries=server_side_retries - 1,
-                                **extra_session_req_kwargs)
+                            'GET request, retrying after %d seconds. Retries '
+                            'left %d.',
+                            e, self._server_side_retries_delay,
+                            server_side_retries_left)
+                time.sleep(self._server_side_retries_delay)
+                server_side_retries_left -= 1
+                return self._op(
+                    method, path, data=data, headers=headers,
+                    blocking=blocking, timeout=timeout,
+                    server_side_retries_left=server_side_retries_left,
+                    **extra_session_req_kwargs)
+            else:
+                raise
 
         if blocking and response.status_code == 202:
             if not response.headers.get('Location'):
@@ -268,13 +299,14 @@ class Connector(object):
                         blocking=blocking, timeout=timeout,
                         **extra_session_req_kwargs)
 
-    def patch(self, path='', data=None, headers=None, blocking=False,
-              timeout=60, **extra_session_req_kwargs):
-        """HTTP PATCH method.
+    def _etag_handler(self, path='', data=None, headers=None, etag=None,
+                      blocking=False, timeout=60, **extra_session_req_kwargs):
+        """eTag handler containing workarounds for PATCH requests with eTags.
 
         :param path: Optional sub-URI path to the resource.
         :param data: Optional JSON data.
         :param headers: Optional dictionary of headers.
+        :param etag: eTag string.
         :param blocking: Whether to block for asynchronous operations.
         :param timeout: Max time in seconds to wait for blocking async call.
         :param extra_session_req_kwargs: Optional keyword argument to pass
@@ -284,9 +316,89 @@ class Connector(object):
         :raises: ConnectionError
         :raises: HTTPError
         """
-        return self._op('PATCH', path, data=data, headers=headers,
-                        blocking=blocking, timeout=timeout,
-                        **extra_session_req_kwargs)
+        if etag is not None:
+            if not headers:
+                headers = {}
+            headers['If-Match'] = etag
+        try:
+            response = self._op('PATCH', path, data=data,
+                                headers=headers,
+                                blocking=blocking, timeout=timeout,
+                                **extra_session_req_kwargs)
+        except exceptions.HTTPError as resp:
+            LOG.warning("Initial request with eTag failed: %s",
+                        resp)
+            if resp.status_code == http_client.PRECONDITION_FAILED:
+                # NOTE(janders) if there was no eTag provided AND the response
+                # is HTTP 412 Precondition Failed, raise the exception
+                if not etag:
+                    raise
+                # checking for weak eTag
+                pattern = re.compile(r'^(W\/)("\w*")$')
+                match = pattern.match(etag)
+                if match:
+                    LOG.info("Weak eTag provided with original request to "
+                             "%s. Attempting to conversion to strong eTag "
+                             "and re-trying.", path)
+                    # trying weak eTag converted to strong
+                    headers['If-Match'] = match.group(2)
+                    try:
+                        response = self._op('PATCH', path, data=data,
+                                            headers=headers,
+                                            blocking=blocking,
+                                            timeout=timeout,
+                                            **extra_session_req_kwargs)
+                    except exceptions.HTTPError as resp:
+                        if (resp.status_code == http_client.
+                           PRECONDITION_FAILED):
+                            LOG.warning("Request to %s with weak eTag "
+                                        "converted to strong eTag also "
+                                        "failed. Making the final attempt "
+                                        "with no eTag specified.", path)
+                        response = None
+                    if response:
+                        return response
+                else:
+                    # eTag is strong, if it failed the only other thing
+                    # to try is removing it entirely
+                    # info
+                    LOG.warning("Strong eTag provided - retrying request to "
+                                "%s with eTag removed.", path)
+                del headers['If-Match']
+                try:
+                    response = self._op('PATCH', path, data=data,
+                                        headers=headers,
+                                        blocking=blocking, timeout=timeout,
+                                        **extra_session_req_kwargs)
+                except exceptions.HTTPError as resp:
+                    LOG.error("Final re-try with eTag removed has failed, "
+                              "raising exception %s", resp)
+                    raise
+            else:
+                raise
+        return response
+
+    def patch(self, path='', data=None, headers=None, etag=None,
+              blocking=False, timeout=60, **extra_session_req_kwargs):
+        """HTTP PATCH method.
+
+        :param path: Optional sub-URI path to the resource.
+        :param data: Optional JSON data.
+        :param headers: Optional dictionary of headers.
+        :param etag: Optional eTag string.
+        :param blocking: Whether to block for asynchronous operations.
+        :param timeout: Max time in seconds to wait for blocking async call.
+        :param extra_session_req_kwargs: Optional keyword argument to pass
+         requests library arguments which would pass on to requests session
+         object.
+        :returns: The response object from the requests library.
+        :raises: ConnectionError
+        :raises: HTTPError
+        """
+        return self._etag_handler(path, data,
+                                  headers, etag,
+                                  blocking, timeout,
+                                  **extra_session_req_kwargs)
 
     def put(self, path='', data=None, headers=None, blocking=False,
             timeout=60, **extra_session_req_kwargs):
