@@ -17,8 +17,10 @@ from http import client as http_client
 
 from sushy import exceptions
 from sushy.resources import base
+from sushy.resources.certificateservice import certificate
 from sushy.resources import common
 from sushy.resources.manager import constants as mgr_cons
+from sushy import utils
 
 
 class ActionsField(base.CompositeField):
@@ -76,6 +78,8 @@ class VirtualMedia(base.ResourceBase):
     _actions = ActionsField('Actions')
     """Insert/eject action for virtual media"""
 
+    _certificates_path = base.Field(['Certificates', '@odata.id'])
+
     def _get_insert_media_uri(self):
         insert_media = self._actions.insert_media if self._actions else None
         use_patch = False
@@ -102,6 +106,17 @@ class VirtualMedia(base.ResourceBase):
             eject_uri = eject_media.target_uri
         return eject_uri, use_patch
 
+    def is_transfer_protocol_required(self, error=None):
+        """Check the response code and body and in case of failure
+
+        Try to determine if it happened due to missing TransferProtocolType.
+        """
+        return (
+            (error.code.endswith(".ActionParameterMissing")
+             or error.code.endswith(".PropertyMissing"))
+            and "#/TransferProtocolType" in error.related_properties
+        )
+
     def insert_media(self, image, inserted=True, write_protected=True,
                      username=None, password=None, transfer_method=None):
         """Attach remote media to virtual media
@@ -157,7 +172,19 @@ class VirtualMedia(base.ResourceBase):
                 payload['Inserted'] = False
             if not write_protected:
                 payload['WriteProtected'] = False
-            self._conn.post(target_uri, data=payload)
+            # NOTE(janders) attempting to detect whether attachment failure is
+            # due to absence of TransferProtocolType param and if so adding it
+            try:
+                self._conn.post(target_uri, data=payload)
+            except exceptions.HTTPError as error:
+                if self.is_transfer_protocol_required(error):
+                    if payload['Image'].startswith('https://'):
+                        payload['TransferProtocolType'] = "HTTPS"
+                    elif payload['Image'].startswith('http://'):
+                        payload['TransferProtocolType'] = "HTTP"
+                    self._conn.post(target_uri, data=payload)
+                else:
+                    raise
         self.invalidate()
 
     def eject_media(self):
@@ -200,6 +227,20 @@ class VirtualMedia(base.ResourceBase):
                          data={'VerifyCertificate': verify_certificate})
         self.invalidate()
 
+    @property
+    @utils.cache_it
+    def certificates(self):
+        """Get the collection of certificates for this device."""
+        if not self._certificates_path:
+            raise exceptions.MissingAttributeError(
+                attribute='Certificates/@odata.id',
+                resource=self._path)
+
+        return certificate.CertificateCollection(
+            self._conn, self._certificates_path,
+            redfish_version=self.redfish_version,
+            registries=self.registries, root=self.root)
+
 
 class VirtualMediaCollection(base.ResourceCollectionBase):
     """A collection of virtual media attached to a Manager"""
