@@ -164,7 +164,8 @@ class ConnectorOpTestCase(base.TestCase):
         self.auth = mock_auth
         super(ConnectorOpTestCase, self).setUp()
         self.conn = connector.Connector(
-            'http://foo.bar:1234', verify=True)
+            'http://foo.bar:1234', verify=True,
+            server_side_retries=10, server_side_retries_delay=3)
         self.conn._auth = mock_auth
         self.data = {'fake': 'data'}
         self.headers = {'X-Fake': 'header'}
@@ -177,7 +178,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.conn._op('GET', path='fake/path', headers=self.headers)
         self.request.assert_called_once_with(
             'GET', 'http://foo.bar:1234/fake/path',
-            headers=self.headers, json=None)
+            headers=self.headers, json=None, verify=True)
 
     def test_response_callback(self):
         mock_response_callback = mock.MagicMock()
@@ -191,21 +192,22 @@ class ConnectorOpTestCase(base.TestCase):
                       allow_redirects=False)
         self.request.assert_called_once_with(
             'GET', 'http://foo.bar:1234/fake/path',
-            headers=self.headers, json=None, allow_redirects=False)
+            headers=self.headers, json=None, allow_redirects=False,
+            verify=True)
 
     def test_ok_post(self):
         self.conn._op('POST', path='fake/path', data=self.data.copy(),
                       headers=self.headers)
         self.request.assert_called_once_with(
             'POST', 'http://foo.bar:1234/fake/path',
-            json=self.data, headers=self.headers)
+            json=self.data, headers=self.headers, verify=True)
 
     def test_ok_put(self):
         self.conn._op('PUT', path='fake/path', data=self.data.copy(),
                       headers=self.headers)
         self.request.assert_called_once_with(
             'PUT', 'http://foo.bar:1234/fake/path',
-            json=self.data, headers=self.headers)
+            json=self.data, headers=self.headers, verify=True)
 
     def test_ok_delete(self):
         expected_headers = self.headers.copy()
@@ -213,7 +215,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.conn._op('DELETE', path='fake/path', headers=self.headers.copy())
         self.request.assert_called_once_with(
             'DELETE', 'http://foo.bar:1234/fake/path',
-            headers=expected_headers, json=None)
+            headers=expected_headers, json=None, verify=True)
 
     def test_ok_post_with_session(self):
         self.conn._session.headers = {}
@@ -225,7 +227,7 @@ class ConnectorOpTestCase(base.TestCase):
                       data=self.data)
         self.request.assert_called_once_with(
             'POST', 'http://foo.bar:1234/fake/path',
-            json=self.data, headers=expected_headers)
+            json=self.data, headers=expected_headers, verify=True)
         self.assertEqual(self.conn._session.headers,
                          {'X-Auth-Token': 'asdf1234'})
 
@@ -238,7 +240,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.conn._op('GET', path=path, headers=headers)
         self.request.assert_called_once_with(
             'GET', 'http://foo.bar:1234' + path,
-            headers=expected_headers, json=None)
+            headers=expected_headers, json=None, verify=True)
 
     def test_odata_version_header_redfish_no_headers(self):
         path = '/redfish/v1/bar'
@@ -246,7 +248,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.conn._op('GET', path=path)
         self.request.assert_called_once_with(
             'GET', 'http://foo.bar:1234' + path,
-            headers=expected_headers, json=None)
+            headers=expected_headers, json=None, verify=True)
 
     def test_odata_version_header_redfish_existing_header(self):
         path = '/redfish/v1/foo'
@@ -255,7 +257,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.conn._op('GET', path=path, headers=headers)
         self.request.assert_called_once_with(
             'GET', 'http://foo.bar:1234' + path,
-            headers=expected_headers, json=None)
+            headers=expected_headers, json=None, verify=True)
 
     def test_timed_out_session_unable_to_create_session(self):
         self.conn._auth.can_refresh_session.return_value = False
@@ -361,6 +363,7 @@ class ConnectorOpTestCase(base.TestCase):
         self.assertEqual(http_client.BAD_REQUEST, exc.status_code)
         self.assertIsNotNone(exc.body)
         self.assertIn('body submitted was malformed JSON', exc.detail)
+        self.assertEqual(len(exc.extended_info), 3, exc.extended_info)
 
     def test_known_http_error_nonlist_ext_info(self):
         self.request.return_value.status_code =\
@@ -376,6 +379,8 @@ class ConnectorOpTestCase(base.TestCase):
         self.assertEqual(http_client.UNSUPPORTED_MEDIA_TYPE, exc.status_code)
         self.assertIsNotNone(exc.body)
         self.assertIn('See Resolution for information', exc.detail)
+        self.assertIn('unsupported media type',
+                      exc.extended_info['Resolution'])
 
     @mock.patch('time.sleep', autospec=True)
     def test_not_found_error(self, mock_sleep):
@@ -401,8 +406,44 @@ class ConnectorOpTestCase(base.TestCase):
             self.conn._op('GET', 'http://foo.bar')
         exc = cm.exception
         self.assertEqual(http_client.INTERNAL_SERVER_ERROR, exc.status_code)
-        self.assertEqual(5, mock_sleep.call_count)
-        self.assertEqual(6, self.request.call_count)
+        self.assertEqual(10, mock_sleep.call_count)
+        self.assertEqual(11, self.request.call_count)
+
+    @mock.patch('time.sleep', autospec=True)
+    def test_op_retry_on_server_500_sys518(self, mock_sleep):
+        response_info = {"error": {"@Message.ExtendedInfo": [
+            {'MessageId': 'IDRAC.2.7.SYS518'}]}}
+        mock_error = mock.Mock()
+        mock_error.status_code = 500
+        mock_error.json.return_value = response_info
+        self.request.return_value.status_code = (
+            http_client.INTERNAL_SERVER_ERROR)
+        self.request.return_value.json.side_effect =\
+            exceptions.ServerSideError(
+                method='DELETE', url='http://foo.bar', response=mock_error)
+
+        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
+                          'http://foo.bar')
+        self.assertEqual(10, mock_sleep.call_count)
+        self.assertEqual(11, self.request.call_count)
+
+    @mock.patch('time.sleep', autospec=True)
+    def test_op_retry_on_server_500_other_than_sys518(self, mock_sleep):
+        response_info = {"error": {"@Message.ExtendedInfo": [
+            {'MessageId': 'IDRAC.2.7.SYS999'}]}}
+        mock_error = mock.Mock()
+        mock_error.status_code = 500
+        mock_error.json.return_value = response_info
+        self.request.return_value.status_code = (
+            http_client.INTERNAL_SERVER_ERROR)
+        self.request.return_value.json.side_effect =\
+            exceptions.ServerSideError(
+                method='DELETE', url='http://foo.bar', response=mock_error)
+
+        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
+                          'http://foo.bar')
+        self.assertEqual(0, mock_sleep.call_count)
+        self.assertEqual(1, self.request.call_count)
 
     def test_access_error(self):
         self.conn._auth = None
@@ -503,7 +544,7 @@ class ConnectorOpTestCase(base.TestCase):
             'this is expected prior to authentication', 'HTTP GET '
             'http://redfish/v1/SessionService returned code '
             '%s. unknown error Extended information: '
-            'none' % http_client.FORBIDDEN)
+            'None' % http_client.FORBIDDEN)
         self.assertEqual(http_client.FORBIDDEN, exc.status_code)
 
     def test_blocking_no_location_header(self):
@@ -539,10 +580,10 @@ class ConnectorOpTestCase(base.TestCase):
     def test_access_error_basic_auth(self, mock_log):
         self.conn._auth.can_refresh_session.return_value = False
         self.conn._session.auth = ('foo', 'bar')
-        self.request.return_value.status_code = http_client.FORBIDDEN
+        self.request.return_value.status_code = 403
         mock_response = mock.Mock()
         mock_response.json.side_effect = ValueError('no json')
-        mock_response.status_code = http_client.FORBIDDEN
+        mock_response.status_code = 403
         self.request.return_value.json.side_effect = ValueError('no json')
 
         with self.assertRaisesRegex(exceptions.AccessError,
@@ -551,12 +592,13 @@ class ConnectorOpTestCase(base.TestCase):
         exc = cm.exception
         self.assertEqual(http_client.FORBIDDEN, exc.status_code)
         self.conn._auth.authenticate.assert_not_called()
+        error_msg = (f'HTTP GET http://redfish/v1/SessionService '
+                     f'returned code {exc.status_code}. unknown error '
+                     f'Extended information: None')
         mock_log.assert_called_once_with(
             "We have encountered an AccessError when using 'basic' "
             "authentication. %(err)s",
-            {'err': 'HTTP GET http://redfish/v1/SessionService '
-                    'returned code HTTPStatus.FORBIDDEN. unknown error '
-                    'Extended information: none'}
+            {'err': error_msg}
         )
 
     def test__op_raises_connection_error(self):
@@ -577,3 +619,175 @@ class ConnectorOpTestCase(base.TestCase):
                               'POST',
                               'http://foo.bar',
                               blocking=True)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_strong_etag_ok(self, mock__op):
+        self.headers['If-Match'] = '"3d7b8a7360bf2941d"'
+        response = mock.MagicMock(spec=requests.Response)
+        response.status_code = http_client.OK
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        mock__op.return_value = response
+        self.conn._etag_handler(path='/redfish/v1/Systems/1',
+                                headers=self.headers, data=data,
+                                etag=self.headers['If-Match'],
+                                blocking=False, timeout=60)
+        mock__op.assert_called_once_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header',
+                     'If-Match': '"3d7b8a7360bf2941d"'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_weak_etag_ok(self, mock__op):
+        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
+        response = mock.MagicMock(spec=requests.Response)
+        response.status_code = http_client.OK
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        mock__op.return_value = response
+        self.conn._etag_handler(path='/redfish/v1/Systems/1',
+                                headers=self.headers, data=data,
+                                etag=self.headers['If-Match'],
+                                blocking=False, timeout=60)
+        mock__op.assert_called_once_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header',
+                     'If-Match': 'W/"3d7b8a7360bf2941d"'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_no_etag_ok(self, mock__op):
+        response = mock.MagicMock(spec=requests.Response)
+        response.status_code = http_client.OK
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        mock__op.return_value = response
+        self.conn._etag_handler(path='/redfish/v1/Systems/1',
+                                headers=self.headers, data=data,
+                                blocking=False, timeout=60)
+        mock__op.assert_called_once_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_weak_etag_fallback_to_strong(self, mock__op):
+        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
+        target_uri = '/redfish/v1/Systems/1'
+        response_412 = mock.MagicMock(spec=requests.Response)
+        response_412.status_code = http_client.PRECONDITION_FAILED
+        response_200 = mock.MagicMock(spec=requests.Response)
+        response_200.status_code = http_client.OK
+        mock__op.side_effect = [
+            exceptions.HTTPError(method='PATCH', url=target_uri,
+                                 response=response_412),
+            response_200]
+
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        self.conn._etag_handler(path=target_uri, headers=self.headers,
+                                data=data, etag=self.headers['If-Match'],
+                                blocking=False, timeout=60)
+        mock__op.assert_called_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header',
+                     'If-Match': '"3d7b8a7360bf2941d"'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_weak_etag_fallback_to_no_etag(self, mock__op):
+        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
+        target_uri = '/redfish/v1/Systems/1'
+        response_412 = mock.MagicMock(spec=requests.Response)
+        response_412.status_code = http_client.PRECONDITION_FAILED
+        response_200 = mock.MagicMock(spec=requests.Response)
+        response_200.status_code = http_client.OK
+        mock__op.side_effect = [
+            exceptions.HTTPError(method='PATCH', url=target_uri,
+                                 response=response_412),
+            exceptions.HTTPError(method='PATCH', url=target_uri,
+                                 response=response_412),
+            response_200]
+
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        self.conn._etag_handler(path=target_uri, headers=self.headers,
+                                data=data, etag=self.headers['If-Match'],
+                                blocking=False, timeout=60)
+        mock__op.assert_called_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_strong_etag_fallback_to_no_etag(self, mock__op):
+        self.headers['If-Match'] = '"3d7b8a7360bf2941d"'
+        target_uri = '/redfish/v1/Systems/1'
+        response_412 = mock.MagicMock(spec=requests.Response)
+        response_412.status_code = http_client.PRECONDITION_FAILED
+        response_200 = mock.MagicMock(spec=requests.Response)
+        response_200.status_code = http_client.OK
+        mock__op.side_effect = [
+            exceptions.HTTPError(method='PATCH', url=target_uri,
+                                 response=response_412),
+            response_200]
+
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        self.conn._etag_handler(path=target_uri, headers=self.headers,
+                                data=data, etag=self.headers['If-Match'],
+                                blocking=False, timeout=60)
+        mock__op.assert_called_with(
+            self.conn,
+            "PATCH",
+            '/redfish/v1/Systems/1',
+            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
+                           'BootSourceOverrideEnabled': 'Once'}},
+            headers={'X-Fake': 'header'},
+            blocking=False,
+            timeout=60)
+
+    @mock.patch.object(connector.Connector, '_op', autospec=True)
+    def test_etag_handler_fail_other_exception(self, mock__op):
+        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
+        target_uri = '/redfish/v1/Systems/1'
+        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
+                         'BootSourceOverrideEnabled': 'Once'}}
+        response = mock.MagicMock(spec=requests.Response)
+        response.status_code = http_client.FORBIDDEN
+        response.message = "boom"
+        mock__op.return_value = response
+        mock__op.side_effect = \
+            exceptions.HTTPError(method='PATCH',
+                                 url=target_uri,
+                                 response=response)
+        self.assertRaises(exceptions.HTTPError, self.conn._etag_handler,
+                          'PATCH', target_uri, data,
+                          self.headers,
+                          blocking=False, timeout=60)
